/******************************************************************************
 * Copyright 2021 TypeFox GmbH
 * This program and the accompanying materials are made available under the
 * terms of the MIT License, which is available in the project root.
 ******************************************************************************/
import type { Connection, TextDocuments } from 'vscode-languageserver';
import type { TextDocument } from 'vscode-languageserver-textdocument';
import type { Grammar } from './grammar/generated/ast.js';
import type { GrammarConfig } from './grammar/grammar-config.js';
import type { LanguageMetaData } from './grammar/language-meta-data.js';
import type { ExecuteCommandHandler } from './lsp/execute-command-handler.js';
import type { CodeActionProvider } from './lsp/code-action.js';
import type { CompletionProvider } from './lsp/completion/completion-provider.js';
import type { DocumentHighlightProvider } from './lsp/document-highlight-provider.js';
import type { DocumentSymbolProvider } from './lsp/document-symbol-provider.js';
import type { FoldingRangeProvider } from './lsp/folding-range-provider.js';
import type { Formatter } from './lsp/formatter.js';
import type { DefinitionProvider } from './lsp/definition-provider.js';
import type { HoverProvider } from './lsp/hover-provider.js';
import type { ReferencesProvider } from './lsp/references-provider.js';
import type { RenameProvider } from './lsp/rename-provider.js';
import type { SemanticTokenProvider } from './lsp/semantic-token-provider.js';
import type { LangiumCompletionParser, LangiumParser } from './parser/langium-parser.js';
import type { IParserConfig } from './parser/parser-config.js';
import type { TokenBuilder } from './parser/token-builder.js';
import type { Lexer } from './parser/lexer.js';
import type { ValueConverter } from './parser/value-converter.js';
import type { Linker } from './references/linker.js';
import type { NameProvider } from './references/name-provider.js';
import type { References } from './references/references.js';
import type { ScopeComputation } from './references/scope-computation.js';
import type { ScopeProvider } from './references/scope-provider.js';
import type { JsonSerializer } from './serializer/json-serializer.js';
import type { ServiceRegistry } from './service-registry.js';
import type { AstReflection } from './syntax-tree.js';
import type { MutexLock } from './utils/promise-util.js';
import type { DocumentValidator } from './validation/document-validator.js';
import type { ValidationRegistry } from './validation/validation-registry.js';
import type { AstNodeDescriptionProvider, ReferenceDescriptionProvider } from './workspace/ast-descriptions.js';
import type { AstNodeLocator } from './workspace/ast-node-locator.js';
import type { ConfigurationProvider } from './workspace/configuration.js';
import type { DocumentBuilder } from './workspace/document-builder.js';
import type { LangiumDocumentFactory, LangiumDocuments } from './workspace/documents.js';
import type { FileSystemProvider } from './workspace/file-system-provider.js';
import type { IndexManager } from './workspace/index-manager.js';
import type { WorkspaceManager } from './workspace/workspace-manager.js';
import type { LanguageServer } from './lsp/language-server.js';
import type { SignatureHelpProvider } from './lsp/signature-help-provider.js';
import type { TypeDefinitionProvider } from './lsp/type-provider.js';
import type { ImplementationProvider } from './lsp/implementation-provider.js';
import type { CallHierarchyProvider } from './lsp/call-hierarchy-provider.js';
import type { DocumentLinkProvider } from './lsp/document-link-provider.js';
import type { CodeLensProvider } from './lsp/code-lens-provider.js';
import type { DeclarationProvider } from './lsp/declaration-provider.js';
import type { DocumentationProvider } from './documentation/documentation-provider.js';
import type { InlayHintProvider } from './lsp/inlay-hint-provider.js';
import type { CommentProvider } from './documentation/comment-provider.js';
import type { WorkspaceSymbolProvider } from './lsp/workspace-symbol-provider.js';
import type { NodeKindProvider } from './lsp/node-kind-provider.js';
import type { FuzzyMatcher } from './lsp/fuzzy-matcher.js';
import type { IParserErrorMessageProvider } from 'chevrotain';
/**
 * The services generated by `langium-cli` for a single language. These are derived from the
 * grammar definition and the language configuration.
 */
export type LangiumGeneratedServices = {
    Grammar: Grammar;
    LanguageMetaData: LanguageMetaData;
    parser: {
        ParserConfig?: IParserConfig;
    };
};
/**
 * Services related to the Language Server Protocol (LSP).
 */
export type LangiumLspServices = {
    CompletionProvider?: CompletionProvider;
    DocumentHighlightProvider?: DocumentHighlightProvider;
    DocumentSymbolProvider?: DocumentSymbolProvider;
    HoverProvider?: HoverProvider;
    FoldingRangeProvider?: FoldingRangeProvider;
    DefinitionProvider?: DefinitionProvider;
    TypeProvider?: TypeDefinitionProvider;
    ImplementationProvider?: ImplementationProvider;
    ReferencesProvider?: ReferencesProvider;
    CodeActionProvider?: CodeActionProvider;
    SemanticTokenProvider?: SemanticTokenProvider;
    RenameProvider?: RenameProvider;
    Formatter?: Formatter;
    SignatureHelp?: SignatureHelpProvider;
    CallHierarchyProvider?: CallHierarchyProvider;
    DeclarationProvider?: DeclarationProvider;
    InlayHintProvider?: InlayHintProvider;
    CodeLensProvider?: CodeLensProvider;
    DocumentLinkProvider?: DocumentLinkProvider;
};
/**
 * Services for a single language where Langium provides default implementations.
 */
export type LangiumDefaultServices = {
    lsp: LangiumLspServices;
    parser: {
        GrammarConfig: GrammarConfig;
        ValueConverter: ValueConverter;
        LangiumParser: LangiumParser;
        ParserErrorMessageProvider: IParserErrorMessageProvider;
        CompletionParser: LangiumCompletionParser;
        TokenBuilder: TokenBuilder;
        Lexer: Lexer;
    };
    documentation: {
        CommentProvider: CommentProvider;
        DocumentationProvider: DocumentationProvider;
    };
    references: {
        Linker: Linker;
        NameProvider: NameProvider;
        References: References;
        ScopeProvider: ScopeProvider;
        ScopeComputation: ScopeComputation;
    };
    serializer: {
        JsonSerializer: JsonSerializer;
    };
    validation: {
        DocumentValidator: DocumentValidator;
        ValidationRegistry: ValidationRegistry;
    };
    workspace: {
        AstNodeLocator: AstNodeLocator;
        AstNodeDescriptionProvider: AstNodeDescriptionProvider;
        ReferenceDescriptionProvider: ReferenceDescriptionProvider;
    };
    shared: LangiumSharedServices;
};
/**
 * The full set of services available for a language. These are either generated by `langium-cli`
 * or provided as default implementations.
 */
export type LangiumServices = LangiumGeneratedServices & LangiumDefaultServices;
/**
 * The services generated by `langium-cli` that are shared between multiple languages. These are
 * derived from the grammar definition.
 */
export type LangiumGeneratedSharedServices = {
    AstReflection: AstReflection;
};
/**
 * Services shared between multiple languages where Langium provides default implementations.
 */
export type LangiumDefaultSharedServices = {
    ServiceRegistry: ServiceRegistry;
    lsp: {
        Connection?: Connection;
        ExecuteCommandHandler?: ExecuteCommandHandler;
        WorkspaceSymbolProvider?: WorkspaceSymbolProvider;
        NodeKindProvider: NodeKindProvider;
        FuzzyMatcher: FuzzyMatcher;
        LanguageServer: LanguageServer;
    };
    workspace: {
        DocumentBuilder: DocumentBuilder;
        IndexManager: IndexManager;
        LangiumDocuments: LangiumDocuments;
        LangiumDocumentFactory: LangiumDocumentFactory;
        TextDocuments: TextDocuments<TextDocument>;
        WorkspaceManager: WorkspaceManager;
        FileSystemProvider: FileSystemProvider;
        MutexLock: MutexLock;
        ConfigurationProvider: ConfigurationProvider;
    };
};
/**
 * The shared services are a set of services that are used by every language within a Langium project.
 * This is necessary to enable features like cross references across different languages.
 */
export type LangiumSharedServices = LangiumDefaultSharedServices & LangiumGeneratedSharedServices;
/**
 * A deep partial type definition for services. We look into T to see whether its type definition contains
 * any methods. If it does, it's one of our services and therefore should not be partialized.
 */
export type DeepPartial<T> = T[keyof T] extends Function ? T : {
    [P in keyof T]?: DeepPartial<T[P]>;
};
/**
 * Language-specific services to be partially overridden via dependency injection.
 */
export type PartialLangiumServices = DeepPartial<LangiumServices>;
/**
 * Shared services to be partially overridden via dependency injection.
 */
export type PartialLangiumSharedServices = DeepPartial<LangiumSharedServices>;
//# sourceMappingURL=services.d.ts.map